from time import time
import threading
from functools import partial
from django.utils.functional import cached_property
from django.core.cache import cache
try:
    import asyncio
except (ImportError, SyntaxError):
    asyncio = None


__all__ = [
    "cached_property", 
    "cached_property_with_ttl", 
    "memoize", 
    'cached',
    "cache_get_or_set"
]


class memoize(object):
    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated.
    """
    def __init__(self, func):
        self.func = func
        self.cache = {}
    def __call__(self, *args):
        try:
            return self.cache[args]
        except KeyError:
            value = self.func(*args)
            self.cache[args] = value
            return value
        except TypeError:
             # uncachable -- for instance, passing a list as an argument.
             # Better to not cache than to blow up entirely.
            return self.func(*args)
    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__
    def __get__(self, obj, objtype):
        """Support instance methods."""
        fn = partial(self.__call__, obj)
        fn.reset = self._reset
        return fn
    def _reset(self):
        self.cache = {}

class cached(memoize): pass


def cache_get_or_set(key, value, ttl):
    return cache.get_or_set(key, value, ttl)


class cached_property:
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance.

    Optional ``name`` argument allows you to make cached properties of other
    methods. (e.g.  url = cached_property(get_absolute_url, name='url') )
    """
    def __init__(self, func, name=None):
        self.func = func
        self.__doc__ = getattr(func, '__doc__')
        self.name = name or func.__name__

    def __get__(self, instance, cls=None):
        """
        Call the function and put the return value in instance.__dict__ so that
        subsequent attribute access on the instance returns the cached value
        instead of calling cached_property.__get__().
        """
        if instance is None:
            return self
        res = instance.__dict__[self.name] = self.func(instance)
        return res


# class cached_property(object):
#     """
#     A property that is only computed once per instance and then replaces itself
#     with an ordinary attribute. Deleting the attribute resets the property.
#     Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76
#     """  # noqa

#     def __init__(self, func):
#         self.__doc__ = getattr(func, "__doc__")
#         self.func = func

#     def __get__(self, obj, cls):
#         if obj is None:
#             return self

#         if asyncio and asyncio.iscoroutinefunction(self.func):
#             return self._wrap_in_coroutine(obj)

#         value = obj.__dict__[self.func.__name__] = self.func(obj)
#         return value

#     def _wrap_in_coroutine(self, obj):

#         @asyncio.coroutine
#         def wrapper():
#             future = asyncio.ensure_future(self.func(obj))
#             obj.__dict__[self.func.__name__] = future
#             return future

#         return wrapper()


class threaded_cached_property(object):
    """
    A cached_property version for use in environments where multiple threads
    might concurrently try to access the property.
    """

    def __init__(self, func):
        self.__doc__ = getattr(func, "__doc__")
        self.func = func
        self.lock = threading.RLock()

    def __get__(self, obj, cls):
        if obj is None:
            return self

        obj_dict = obj.__dict__
        name = self.func.__name__
        with self.lock:
            try:
                # check if the value was computed before the lock was acquired
                return obj_dict[name]

            except KeyError:
                # if not, do the calculation and release the lock
                return obj_dict.setdefault(name, self.func(obj))


class cached_property_with_ttl(object):
    """
    A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Setting the ttl to a number expresses how long
    the property will last before being timed out.
    """

    def __init__(self, ttl=None):
        if callable(ttl):
            func = ttl
            ttl = None
        else:
            func = None
        self.ttl = ttl
        self._prepare_func(func)

    def __call__(self, func):
        self._prepare_func(func)
        return self

    def __get__(self, obj, cls):
        if obj is None:
            return self

        now = time()
        obj_dict = obj.__dict__
        name = self.__name__
        try:
            value, last_updated = obj_dict[name]
        except KeyError:
            pass
        else:
            ttl_expired = self.ttl and self.ttl < now - last_updated
            if not ttl_expired:
                return value

        value = self.func(obj)
        obj_dict[name] = (value, now)
        return value

    def __delete__(self, obj):
        obj.__dict__.pop(self.__name__, None)

    def __set__(self, obj, value):
        obj.__dict__[self.__name__] = (value, time())

    def _prepare_func(self, func):
        self.func = func
        if func:
            self.__doc__ = func.__doc__
            self.__name__ = func.__name__
            self.__module__ = func.__module__


# Aliases to make cached_property_with_ttl easier to use
cached_property_ttl = cached_property_with_ttl
timed_cached_property = cached_property_with_ttl


class threaded_cached_property_with_ttl(cached_property_with_ttl):
    """
    A cached_property version for use in environments where multiple threads
    might concurrently try to access the property.
    """

    def __init__(self, ttl=None):
        super(threaded_cached_property_with_ttl, self).__init__(ttl)
        self.lock = threading.RLock()

    def __get__(self, obj, cls):
        with self.lock:
            return super(threaded_cached_property_with_ttl, self).__get__(obj, cls)


# Alias to make threaded_cached_property_with_ttl easier to use
threaded_cached_property_ttl = threaded_cached_property_with_ttl
timed_threaded_cached_property = threaded_cached_property_with_ttl


import datetime
import workon.utils
import workon.conf
from django import forms
from django.db import models
from django.views import generic
from django.shortcuts import render
from django.utils.html import mark_safe
from django.utils import six
from django.utils.text import slugify
from django.forms.fields import CallableChoiceIterator
from django.core.exceptions import FieldDoesNotExist


__all__ = ['ListColumn', 'ListCol', 'ListFilter', 'ListAction', 'List', 'TableList']


    









class ModelListChoiceField(forms.ModelChoiceField):

    empty_label= ''

    pass

class ListChoiceField(forms.ChoiceField):
    def _get_choices(self):
        return self._choices

    def _set_choices(self, value):
        if callable(value):
            value = CallableChoiceIterator(value)
        else:
            value = list(value)
        self._choices = self.widget.choices = [('', ' -- '), ] + list(value)
    choices = property(_get_choices, _set_choices)


class ListMultipleChoiceField(forms.MultipleChoiceField):
    def _get_choices(self):
        return self._choices

    def _set_choices(self, value):
        if callable(value):
            value = CallableChoiceIterator(value)
        else:
            value = list(value)
        self._choices = self.widget.choices = list(value)
    choices = property(_get_choices, _set_choices)
















class ListAction():
    def __init__(self, name, label=None, col=2):
        self.name = name
        self.meta = {
            'class': f'col s12 m{col}',
            'label': label or name,
        }










class ListFilter():
    def __init__(self, name, label=None, col=2, choices=None, classes='', 
        multiple=False, default=None, field_class=None, field_kwargs=None,
        linebreaks=False
    ):
        self.name = name
        field_kwargs = field_kwargs if field_kwargs else {}
        field_kwargs['required'] = False
        field_kwargs['initial'] = default
        field_kwargs['label'] = label
        if not field_class:
            if choices is not None:
                if isinstance(choices, models.QuerySet):
                    field_class = ModelListChoiceField
                    field_kwargs['queryset'] = choices
                    field_kwargs['empty_label'] = ''
                else:
                    if multiple == True:
                        field_class = ListMultipleChoiceField
                    else:
                        field_class = ListChoiceField
                    field_kwargs['choices'] = choices
            else:
                field_class = forms.CharField
        self.form_field_kwargs = field_kwargs
        self.form_field_class = field_class
        self.meta = {
            'class': f'{classes}',
            'label': label if label else name,
            'linebreaks': linebreaks
        }



class ListForm(forms.Form):
    def __init__(self, filters_instances,  *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['page'] = forms.CharField(widget=forms.HiddenInput(), required=False)
        for fi in filters_instances:
            field_kwargs = fi.form_field_kwargs
            self.fields[fi.name] = fi.form_field_class(**field_kwargs)
            self.fields[fi.name].meta = fi.meta
            self.fields[fi.name].filter = fi





class ListColumn():
    def __init__(self, name, label=None, classes='', attrs='', ellipsis=False, col=None, value=None, model=None):
        self.name = name
        self.label = label
        self.classes = classes
        self.attrs = attrs
        self.ellipsis = ellipsis
        self.value = value
        self.model = model

    def __str__(self):
        return str(self.label) if self.label else ''

    @classmethod
    def make_instance(cls, listview, col, **kwargs):
        if isinstance(col, cls):
            col = col

        elif isinstance(col, dict):
            col = cls(**col)

        elif isinstance(col, tuple):
            label = col[-1]
            col = cls.make_instance(listview, col[0], label=col[-1])

        elif isinstance(col, six.string_types):
            col = cls(col, **kwargs)

        if listview.model and not col.label:
            try:
                col.label = listview.model._meta.get_field(col.name).verbose_name
            except FieldDoesNotExist:
                print('FieldDoesNotExist')
                pass
        return col

class ListCol(ListColumn): pass






class List(generic.FormView):

    template_name = workon.conf.LIST_TEMPLATE
    results_template_name = workon.conf.LIST_RESULTS_TEMPLATE
    row_template_name = workon.conf.LIST_ROW_TEMPLATE
    layout_template_name = workon.conf.LIST_LAYOUT_TEMPLATE
    floating_actions_template_name = workon.conf.LIST_FLOATING_ACTIONS_TEMPLATE

    list_id = 'filtered_list'
    list_class = ''
    filters = []
    actions = []
    columns = []
    filters_instances = []
    columns_instances = []
    form_class = None
    form_prefix = 'wlff' # workon_list_filters_form
    model = None
    ajax_results = True

    paginate_by = 50
    paginate_body = 6
    paginate_padding = 2

    create_url = None
    create_method = workon.conf.LIST_ROW_CREATE_METHOD
    update_method = workon.conf.LIST_ROW_UPDATE_METHOD
    view_method = workon.conf.LIST_ROW_VIEW_METHOD
    delete_method = workon.conf.LIST_ROW_DELETE_METHOD
    update_on_double_click_method = workon.conf.LIST_ROW_UPDATE_ON_DOUBLE_CLICK_METHOD

    row_class = ''
    row_style = ''
    row_id = lambda self, obj: f'{obj}_{obj.pk}' if hasattr(obj, 'pk') else str(obj)

    def __get_vomr(self, name):        
        value = getattr(self, f'get_{name}', getattr(self, name, None))
        if workon.utils.is_lambda(value):
            return value()
        elif workon.utils.is_method(value):
            return value()
        else:
            return getattr(self, name, None)

    def __get_vomr_obj(self, name, obj):        
        value = getattr(self, f'get_{name}', getattr(self, name, None))

        ## COMPAT
        if value is None:
            value = getattr(self, f'get_col_{name}', None)
        ## END_COMPAT

        if workon.utils.is_lambda(value):
            return value(obj)
        elif workon.utils.is_method(value):
            return value(obj)
        else:
            return getattr(obj, name, None)

    def __get_vomr_obj_value(self, name, obj):        
        value = getattr(self, f'get_{name}_value', getattr(self, f'{name}_value', None))

        ## COMPAT
        if value is None:
            value = getattr(self, f'get_col_{name}_value', None)
        ## END_COMPAT

        if value is None:
            value = getattr(obj, name, None)

        if workon.utils.is_lambda(value):
            return value(obj)
        elif workon.utils.is_method(value):
            return value(obj)
        else:
            return getattr(obj, name, None)

    def make_filter_instance(self, fi, **kwargs):
        if isinstance(fi, ListFilter):
            fi = fi

        elif isinstance(fi, dict):
            fi = ListFilter(**fi)

        elif isinstance(fi, tuple):
            fi = self.make_filter_instance(fi[0], label=fi[-1])

        elif isinstance(fi, six.string_types):
            name_ftype = fi.split('::', 1)
            ftype = name_ftype[-1]
            name = name_ftype[0]
            if not 'label' in kwargs:
                kwargs['label'] = name
            if len(name_ftype) == 2:
                if ftype == "int":
                    kwargs['field_class'] = forms.IntegerField
                elif ftype in ["bool", "boolean"]:
                    kwargs['field_class'] = forms.BooleanField
                elif ftype in ["hidden"]:
                    kwargs['field_class'] = forms.TextField
                    kwargs['field_kwargs'] = {'widget': forms.HiddenInput()}
                elif ftype in ["date"]:
                    kwargs['field_class'] = workon.forms.DateField
                elif ftype in ["datetime"]:
                    kwargs['field_class'] = workon.forms.DateTimeField
                elif ftype in ["time"]:
                    kwargs['field_class'] = workon.forms.TimeField
            fi = ListFilter(name, **kwargs)
        return fi


    def is_last_object_saved(self, obj):
        return f'{getattr(obj, "__class___", None)}_{getattr(obj, "id", None)}' == self.request.session.get("wls_last_object_saved")

    def get_form(self):

        self.columns_instances = [ ListCol.make_instance(self, col) for col in getattr(self, 'fields', self.columns)]
        self.filters_instances = [ self.make_filter_instance(fi) for fi in self.filters]

        if not self.form_class:
            self.form = ListForm(self.filters_instances, **self.get_form_kwargs())
        else:
            self.form = self.form_class(**self.get_form_kwargs())
        self.form_initialize(self.form)
        return self.form

    def form_initialize(self, form):
        return form

    def get_prefix(self):
        return self.form_prefix
        
    def get_form_kwargs(self):
        prefix = self.get_prefix()
        kwargs = {
            'initial': self.get_initial(),
            'prefix': prefix,
        }
        if self.request.method in ('GET'):
            kwargs.update({
                'initial': {k.lstrip(f'{prefix}-'): v for k,v in self.request.GET.items()},
            })
            if not self.ajax_results:
                kwargs.update({
                    'data': self.request.GET,
                })

        if self.request.method in ('POST', 'PUT'):
            kwargs.update({
                'data': self.request.POST,
                'files': self.request.FILES,
            })
        return kwargs

    def get_extra_actions(self, obj):
        return ''

    # def get_row_id(self, obj):
    #     if hasattr(obj, 'pk'):
    #         return f'{obj}_{obj.pk}'
    #     else:
    #         return str(obj)


    # def get_row_attrs(self, obj):
    #     if workon.conf.LIST_ROW_UPDATE_ON_DOUBLE_CLICK:
    #         update_url = self.get_row_update_url(obj)
    #         if update_url:
    #             return f'''\
    #                         {self.get_row_update_on_double_click_method(obj)}="{update_url}" \
    #                         data-tooltip='{{ 
    #                             "content": "Double-clic pour éditer" ,
    #                             "position": "left" 
    #                         }}' '''
    #     return ''

    def get_template_names(self):
        return self.results_template_name if self.request.is_ajax() else self.template_name

    def get_queryset(self):
        qs = []
        if getattr(self, 'model', None):
            qs = self.model.objects.all()
        return self.filter(qs, self.F)

    def filter(self, qs, filters=None):
        if hasattr(self, 'model'):
            pass
        return qs

    def get_rows(self):
        self.rows = []
        self.queryset = getattr(self, 'queryset', [])
        # data = getattr(self.form, 'cleaned_data', { f.name: '' for f in self.filters_instances })
        for obj in self.queryset:
            row = self.get_row(obj)
            self.rows.append(row)
        return self.rows

    def get_row(self, obj):
        cells = []
        for column_instance in self.columns_instances:
            cells.append(self.render_cell(column_instance, obj))
        return {
            'object': obj,
            'id': self.__get_vomr_obj('row_id', obj),
            'attrs': self.__get_vomr_obj('row_attrs', obj),
            'class': self.__get_vomr_obj('row_class', obj),
            'style': self.__get_vomr_obj('row_style', obj),
            'cells': cells,
            'update_method': self.__get_vomr_obj('row_update_method', obj),
            'delete_method': self.__get_vomr_obj('row_delete_method', obj),
            'view_method': self.__get_vomr_obj('row_view_method', obj),
            'extra_actions': self.get_extra_actions(obj),
        }

    @classmethod
    def render_row(cls, request, obj):
        self = cls()
        self.request = request
        self.get_form()
        return workon.render_content(request, self.__get_vomr('row_template_name'), {
            'row': self.get_row(obj)
        })

    def render_cell(self, column_instance, obj):
        data = dict(
            attrs='',
            classes='',
            column=column_instance
        )
        name = column_instance.name 
        if column_instance.value is not None:
            if workon.is_lambda(column_instance.value):
                value = column_instance.value(self, obj)
            elif workon.is_method(column_instance.value):
                value = column_instance.value(self, obj)
            else:
                value = column_instance.value
        else:
            value = self.__get_vomr_obj_value(name, obj)

        if value is None or value is "":
            value = f'''<i class="icon disabled">block</i>'''
        elif isinstance(value, bool):
            value = f'''<i class="icon {'success' if value else 'error'}">{'check' if value else 'close'}</i>'''
        elif isinstance(value, datetime.date):
            value = value.strftime("%d %b <sup>%Y</sup> <sub>à %H:%M</sub>").replace('<sub>à 00:00</sub>', '')
        elif isinstance(value, six.string_types):
            value = value
        data['value'] = value 

        classes = self.__get_vomr_obj(f'{name}_class', obj)#getattr(self, f'get_col_{name}_class', None)
        if classes:
            data['classes'] += classes

        attrs = self.__get_vomr_obj(f'{name}_attrs', obj)
        if attrs:
            data['attrs'] += attrs

        # data['attrs'] = " ".join([f'{name}="{value}"' for name, value in data['attrs'].items()]) 
        return data

    def get_breadbrumbs(self):
        return []

    def form_valid(self, form):
        self.data = self.F = form.cleaned_data
        self.queryset = self.get_queryset()
        action = self.request.GET.get('_action')
        if action and hasattr(self, f'action_{action}'):
            response = getattr(self, f'action_{action}')(self.queryset)
            if not response:
                self.queryset = self.get_queryset()
            elif isinstance(response, models.QuerySet):
                self.queryset = response
            else:
                return response
        return self.render_valid(form)

    def render_valid(self, form):
        # try:
        self.paginate()
        self.rows = self.get_rows()
        # except:
        #     self.queryset = qs
        return render(self.request, self.get_template_names(), self.get_context_data())   

    def paginate(self):
        page_param = f'{self.form_prefix}-page' if self.form_prefix else 'page'
        page = self.data.get(page_param, self.data.get('page'))
        if not page:
            page = self.request.POST.get(page_param, self.request.POST.get('page'))
        if not page:
            page = self.request.GET.get(page_param, self.request.GET.get('page'))
        self.queryset = workon.utils.DiggPaginator(
            self.queryset, 
            self.paginate_by, 
            body=self.paginate_body, 
            padding=self.paginate_padding
        ).get_queryset_for_page(
            page
        )     

    def form_invalid(self, form):
        return self.form_valid(form)

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)

        ctx['page_param'] = f'{self.form_prefix}-page' if self.form_prefix else 'page'
        ctx['breadcrumbs'] = self.get_breadbrumbs()
        ctx['columns'] = self.columns_instances
        ctx['actions'] = self.actions
        ctx['rows'] = getattr(self, 'rows', [])
        ctx['objects'] = ctx['queryset'] = getattr(self, 'queryset', [])
        ctx['create_method'] = self.__get_vomr('create_method')
        ctx['create_url'] = self.__get_vomr('create_url')
        ctx['list_results_template'] = self.__get_vomr('results_template_name')
        ctx['list_row_template'] = self.__get_vomr('row_template_name')
        ctx['floating_actions_template'] = self.__get_vomr('floating_actions_template_name')
        ctx['layout_template'] = self.__get_vomr('layout_template_name')
        ctx['list_id'] = self.__get_vomr('list_id')
        ctx['list_class'] = self.__get_vomr('list_class')
        ctx['queryset'] = getattr(self, 'queryset', [])
        return ctx

class RowList(List):
    template_name = "workon/views/list/list.html"
    results_template_name = "workon/views/list/_list.html"
    row_template_name = "workon/views/list/_row.html"
    
    row_nums = 12
    row_tag = 'span'

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx['row_nums'] = self.row_nums
        ctx['row_tag'] = self.row_tag
        return ctx


class TableList(List):

    template_name = "workon/views/list/list.html"
    results_template_name = "workon/views/list/_list.html"
    row_template_name = "workon/views/list/_row.html"


from django.db import models
from django import forms
from django.utils.text import capfirst
from workon.forms import IconField as IconFormField, IconInput


__all__ = ['IconField']


class IconField(models.CharField):
    """
    A text field made to accept hexadecimal color value (#FFFFFF)
    with a color picker widget.
    """
    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 7
        super().__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        kwargs['widget'] = IconInput
        return super().formfield(**kwargs)




import random
try:
    from colour import Color
except:
    from ._colour import Color


__all__ = ['Color', 'PColors', "rgb_to_hex", "generate_colors", "random_color"]


class PColors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    GREEN = OKGREEN
    WARNING = '\033[93m'
    YELLOW = WARNING
    FAIL = '\033[91m'
    ERROR = FAIL
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    OK = f'''{BOLD}{OKGREEN}[OK]{ENDC}'''


def rgb_to_hex(rgb_tuple):
    assert len(rgb_tuple) == 3
    denormalized_values = tuple(map(lambda x: 256*x, rgb_tuple))
    return '#%02x%02x%02x' % denormalized_values

def generate_colors(num, from_color='#f7aabc', to_color='#404a58'):
    """
    Generate `num` distinct Hexadecimal colors
    """
    from_color = Color(from_color)
    to_color = Color(to_color)

    if num == 0:
        return []
    elif num == 1:
        return [from_color.hex]
    return list(c.hex for c in from_color.range_to(to_color, num))

def random_color(luminance=0.5):
    import random
    r = lambda: random.randint(0,255)
    return Color('#%02X%02X%02X' % (r(),r(),r()), luminance=luminance)









import json
import datetime
import time
import re
import os
import logging
import warnings
from functools import reduce
from django import forms
from django.conf import settings
from django.core.validators import EMPTY_VALUES
from django.template import Context
from django.forms.widgets import FileInput as OriginalFileInput
from django.template.loader import render_to_string
from django.forms.utils import flatatt
from django.utils.safestring import mark_safe
from django.utils import datetime_safe, formats, six, translation
from django.utils.translation import ugettext_lazy as _, ungettext_lazy


__all__ = [
    'DateField', 
    'DateTimeField', 
    'TimeField', 
    'DateTimeInput', 
    'DateInput', 
    'TimeInput'
]


class DateField(forms.DateField):

    default_error_messages = {
        'invalid': _('Enter a valid date.'),
    }

    def __init__(self, *args, **kwargs):

        # self.input_formats = formats.get_format('DATE_INPUT_FORMATS')
        if 'widget' not in kwargs:
            kwargs['widget'] = DateInput(
                # format="%d/%m/%Y",
                attrs={'placeholder': _('jj/mm/aaaa')})
            # kwargs['input_formats'] = ['%m/%d/%Y']
        super().__init__(*args, **kwargs)


class DateTimeField(forms.DateTimeField):

    def __init__(self, *args, **kwargs):
        # self.input_formats = formats.get_format('DATETIME_INPUT_FORMATS')

        if 'widget' not in kwargs:
            kwargs['widget'] = DateTimeInput(
                # format="%d/%m/%Y %H:%M:%S",
                attrs={'placeholder': _('jj/mm/aaaa')})
            #kwargs['input_formats'] = ['%d/%m/%Y %H:%M']
        super().__init__(*args, **kwargs)

class TimeField(forms.DateTimeField):

    def __init__(self, *args, **kwargs):
        # self.input_formats = formats.get_format('TIME_INPUT_FORMATS')
        if 'widget' not in kwargs:
            kwargs['widget'] = TimeInput(
                # format="%d/%m/%Y %H:%M:%S",
                attrs={'placeholder': _('hh:mm')})
            #kwargs['input_formats'] = ['%d/%m/%Y %H:%M']
        super().__init__(*args, **kwargs)



FORMAT_MAP = {
    'd': r'%d',
    'm': r'%m',
    'Y': r'%Y',
    'H': r'%H',
    'i': r'%M',
    's': r'%S'
}
# FORMAT_MAP_REVERSE = {v:k for k,v in FORMAT_MAP.items()}

def _py_datetime_format_to_js(format_string):
    return (
        format_string is not None and
        reduce(
            lambda format_string, args: format_string.replace(*reversed(args)),
            FORMAT_MAP.items(),
            format_string
        ) or
        None
    )


def _js_datetime_format_to_py(format_string):
    return (
        format_string is not None and
        reduce(
            lambda format_string, args: format_string.replace(*args),
            FORMAT_MAP.items(),
            format_string
        ) or
        None
    )

class BaseDateInput(forms.DateTimeInput):
    # Build a widget which uses the locale datetime format but without seconds.
    # We also use data attributes to pass these formats to the JS datepicker.

    timepicker = False
    # class Media:
    #     css = {
    #         'all': ('contrib/vendors/datetimepicker/jquery.datetimepicker.css',)
    #     }
    #     js = ('contrib/vendors/datetimepicker/jquery.datetimepicker.js',)

    input_type = 'text'
    # @property
    # def media(self):
    #     return forms.Media(
    #         # js=[
    #         #     # 'https://cdnjs.cloudflare.com/ajax/libs/moment.js/2.13.0/locale/fr.js',
    #         #     'contrib/packages/datetime.js'
    #         # ],
    #         # css={'all': ["contrib/vendors/datetimepicker/jquery.datetimepicker.css"]}
    #     )

    def __init__(self, *args, **kwargs):

        include_seconds = kwargs.pop('include_seconds', False)
        datetime = kwargs.get('attrs', {}).get('datetime', None)
        options = kwargs.pop('options', {
            'lang': translation.get_language(),
            'timepicker': self.timepicker,
            'datepicker': self.datepicker,
            'mask': False
        })
        formats_set = {
            datetime and _js_datetime_format_to_py(datetime) or None,
            kwargs.get('format_string', None),
            options.get('format')
        } - {None}  # and this is why (*)
        if len(formats_set) is 0:
            self.format_string =  formats.get_format(self.format_key)[0]
        elif len(formats_set) is 1:
            self.format_string = formats_set.pop()
        else:
            self.format_string = formats_set.pop()
            warnings.warn('format is set more than once', UserWarning)

        self.format_py = _js_datetime_format_to_py(self.format_string)
        self.format_js = _py_datetime_format_to_js(self.format_string)

        if self.format_string:
            options.update({'format': self.format_js})
        options.update({'language': translation.get_language()})

        super().__init__(*args, **kwargs)
        attrs = {
            'data-datetime-widget': json.dumps(options),
            'type': 'text',
            'autocomplete': 'off'
        }
        self.attrs.update(attrs)

    def format_value(self, value):
        return formats.localize_input(value, self.format_string)

    def render(self, name, value, attrs={}, **kwargs):
        renderer = kwargs.get('renderer', None)
        if 'id' not in attrs:
            attrs['id'] = "id_%s" % name
        render = super(BaseDateInput, self).render(name, value, attrs, **kwargs)
        return mark_safe("%s" % (render))


class DateInput(BaseDateInput):
    timepicker = False
    datepicker = True
    format_key = 'DATE_INPUT_FORMATS'


class DateTimeInput(BaseDateInput):
    timepicker = True
    datepicker = True
    format_key = 'DATETIME_INPUT_FORMATS'

class TimeInput(BaseDateInput):
    timepicker = True
    datepicker = False
    format_key = 'TIME_INPUT_FORMATS'